/*
   This ast2infix() method takes an Abstract Syntax Tree
   for Language_2 and converts it into a string that
   contains the infix version of the expression.

   ast2infix() uses knowledge about operator precedence
   (in infix notation) to use the minimun number of
   parentheses.
*/

public class AST2infix
{
   public static String ast2infix(Tree tree)
   {
      String result = "";

      String node = tree.getElement();

      if ( node.equals("prog") )
      {
         result += "Program\n{\n";

         // process each expression
         for (int i = 0; i < tree.degree(); i++)
         {
            result += ast2infix( tree.getSubTree(i), "   ", false );

            if ( tree.getSubTree(i).getElement().equals("fun") )
            {  // give function definitions an extra \n
               result += "\n";
            }

            if ( tree.getSubTree(i).getElement().equals("fun")
              || tree.getSubTree(i).getElement().equals("if")
              || tree.getSubTree(i).getElement().equals("while") )
            {  // the above expressions do not need a semicolon
               result += "\n";
            }
            else // include a semicolon
               result += ";\n";
         }

         result += "}";
      }
      else  // a single expression
      {
         result += ast2infix( tree, "", false );
      }

      return result;
   }

   /**
      This ast2infix() method is essentially
      a in-order traversal of the tree.
   */
   public static String ast2infix(Tree tree, String indent, boolean inline)
   {
      String result = "";

      String node = tree.getElement();

      if ( ! inline )
         result += indent;

      if ( node.equals("lambda") )
      {
         if ( tree.degree() == 2
           && tree.getSubTree(1).depth() < 3 ) // do an inline lambda expression
         {
            // get the number of formal parameters from the lambda expression
            int numberOfParameters = tree.getSubTree(0).degree();
            if (numberOfParameters != 1)
               result += "(";
            // do the list of formal parameters from the lambda expression
            for (int i = 0; i < numberOfParameters; i++)
            {
               String variable = tree.getSubTree(0).getSubTree(i).getElement();
               result += variable;
               if (i < numberOfParameters - 1)
                  result += ", ";
            }
            if (numberOfParameters != 1)
               result += ")";

            // use a Maple like notation
            result += " -> ";
            // get the body of the lambda expression
            result += ast2infix( tree.getSubTree(1), indent, true );
         }
         else  // a C/C++/Java style function body
         {
            result += "(";
            // get the number of formal parameters from the lambda expression
            int numberOfParameters = tree.getSubTree(0).degree();
            // do the list of formal parameters from the lambda expression
            for (int i = 0; i < numberOfParameters; i++)
            {
               String variable = tree.getSubTree(0).getSubTree(i).getElement();
               result += variable;
               if (i < numberOfParameters - 1)
                  result += ", ";
            }
            result += ")\n";
            result += indent + "{\n";
            // process each expression in the lambda expression body
            for(int i = 1; i < tree.degree(); i++)
            {
               result += ast2infix( tree.getSubTree(i), indent + "   ", false );
               if ( tree.getSubTree(i).getElement().equals("if")
                 || tree.getSubTree(i).getElement().equals("while") )
               {  // the above expressions do not need a semicolon
                  result += "\n";
               }
               else
                  result += ";\n";
            }
            result += indent + "}";
         }
      }
      else if ( node.equals("fun") ) // do a better C/C++/Java style function
      {
         // get the function's name
         String functionName = tree.getSubTree(0).getElement();

         // get the lambda expression
         Tree lambdaExp = tree.getSubTree(1);

         // do the name of the function
         result += functionName + "(";
         // get the number of formal parameters from the lambda expression
         int numberOfParameters = lambdaExp.getSubTree(0).degree();
         // do the list of formal parameters from the lambda expression
         for (int i = 0; i < numberOfParameters; i++)
         {
            String variable = lambdaExp.getSubTree(0).getSubTree(i).getElement();
            result += variable;
            if (i < numberOfParameters - 1)
               result += ", ";
         }
         result += ")\n";
         // do the function body from the lambda expression
         // a C/C++/Java style function body
         result += indent + "{\n";
         // process each expression in the lambda expression
         for(int i = 1; i < lambdaExp.degree(); i++)
         {
            result += ast2infix( lambdaExp.getSubTree(i), indent + "   ", false );
            if ( lambdaExp.getSubTree(i).getElement().equals("if")
              || lambdaExp.getSubTree(i).getElement().equals("while") )
            {  // the above expressions do not need a semicolon
               result += "\n";
            }
            else
               result += ";\n";
         }
         result += indent + "}";
      }
      else if ( node.equals("if") )
      {
         result += "if (" + ast2infix( tree.getSubTree(0), indent+"       ", true ) + ")\n";

         if ( tree.getSubTree(1).getElement().equals("while")
            ||tree.getSubTree(1).getElement().equals("if")
            ||tree.getSubTree(1).getElement().equals("lambda") )
         {
            result += ast2infix( tree.getSubTree(1), indent+"   ", inline );
         }
         else
         {
            result += ast2infix( tree.getSubTree(1), indent+"   ", inline );
            result += ";";
         }

         result += "\n" + indent + "else\n";

         if ( tree.getSubTree(2).getElement().equals("while")
            ||tree.getSubTree(2).getElement().equals("if")
            ||tree.getSubTree(1).getElement().equals("lambda") )
         {
            result += ast2infix( tree.getSubTree(2), indent+"   ", inline );
         }
         else
         {
            result += ast2infix( tree.getSubTree(2), indent+"   ", inline ) + ";";
         }
      }
      else if ( node.equals("while") )
      {
         result += "while (" + ast2infix(tree.getSubTree(0), indent+"       ", true) + ")\n";
         result += indent + "{\n";
         // process each expression
         for(int i = 1; i < tree.degree(); i++)
         {
            result += ast2infix( tree.getSubTree(i), indent + "   ", false );
            if ( tree.getSubTree(i).getElement().equals("if")
              || tree.getSubTree(i).getElement().equals("while") )
            {  // the above expressions do not need a semicolon
               result += "\n";
            }
            else
               result += ";\n";
         }
         result += indent + "}";
      }
      else if ( node.equals("var") )
      {
       //indentation += 7 + tree.getSubTree(0).getElement().length();
         result += "var " + tree.getSubTree(0).getElement()
                   + " = " + ast2infix( tree.getSubTree(1), "       "+indent, true );
      }
      else if ( node.equals("set") )
      {
       //indent += 3 + tree.getSubTree(0).getElement().length();
         result += tree.getSubTree(0).getElement()
                   + " = " + ast2infix( tree.getSubTree(1), "   "+indent, true );
      }
      else if ( node.equals("print") )
      {
         result += "print( " + ast2infix( tree.getSubTree(0), indent, true ) + " )";
      }
      else if ( tree.degree() == 0 )  // constant term
      {
         result += tree.getElement();
      }
      else if ( notOperator(node) ) // a function call (apply)
      {
         // do the function value
         if ( tree.getSubTree(0).degree() > 0 )  // if the function value is more
            result += "(";                       // complicated than just a name
         result += ast2infix(tree.getSubTree(0), indent, true);
         if ( tree.getSubTree(0).degree() > 0 )
            result += ")";

         // do the actual parameters
         result += "(";
         for (int i = 1; i < tree.degree(); i++)
         {
            result += ast2infix( tree.getSubTree(i), indent+"   ", true );
            if (i < tree.degree() - 1)
               result += ", ";
         }
         result += ")";
      }
      else if ( tree.degree() == 1 )  // unary operator
      {
         if ( tree.getSubTree(0).degree() >= 2 )
         {
            result += tree.getElement() + "(" + ast2infix( tree.getSubTree(0), indent, true ) + ")";
         }
         else
         {
            result += tree.getElement() + ast2infix( tree.getSubTree(0), indent, true );
         }
      }
      else if ( tree.degree() == 2 )  // binary operator
      {
         int middlePrecedence = precedence( tree.getElement() );
         int leftPrecedence = precedence( tree.getSubTree(0).getElement() );

         if ( leftPrecedence > middlePrecedence )
         {
            result += "(" + ast2infix( tree.getSubTree(0), indent, true ) + ")";
         }
         else
         {
            result += ast2infix( tree.getSubTree(0), indent, true );
         }

         result += " " + tree.getElement() + " ";

         int rightPrecedence = precedence( tree.getSubTree(1).getElement() );

         if ( rightPrecedence >= middlePrecedence )
         {
            result += "(" + ast2infix( tree.getSubTree(1), indent, true ) + ")";
         }
         else
         {
            result += ast2infix( tree.getSubTree(1), indent, true );
         }
      }
      else  // operators with 3 or more operands
      {
         int operatorPrecedence = precedence( tree.getElement() );
         int leftPrecedence = precedence( tree.getSubTree(0).getElement() );

         if ( leftPrecedence > operatorPrecedence )
         {
            result += "(" + ast2infix( tree.getSubTree(0), indent, true ) + ")";
         }
         else
         {
            result += ast2infix( tree.getSubTree(0), indent, true );
         }

         for (int i = 1; i < tree.degree(); i++)
         {
            result += " " + tree.getElement() + " ";

            int rightPrecedence = precedence( tree.getSubTree(i).getElement() );

            if ( rightPrecedence >= operatorPrecedence )
            {
               result += "(" + ast2infix( tree.getSubTree(i), indent, true ) + ")";
            }
            else
            {
               result += ast2infix( tree.getSubTree(i), indent, true );
            }
         }
      }
      return result;
   }


   private static boolean notOperator(String node)
   {
      return ( 0 == precedence(node) );
   }


   // Add the 'var' and 'print' operators to this function.
   // http://www.java-tips.org/java-se-tips/java.lang/what-is-java-operator-precedence.html
   // http://www.cs.princeton.edu/introcs/11precedence/
   private static int precedence(String op)
   {
      int result = 0;  // "highest" precedence (i.e., constants)

      if ( op.equals("!") )
      {
         result = 2;
      }
      else if ( op.equals("^") )
      {
         result = 3;
      }
      else if ( op.equals("*")
             || op.equals("/")
             || op.equals("%") )
      {
         result = 4;
      }
      else if ( op.equals("+")
             || op.equals("-") )
      {
         result = 5;
      }
      else if ( op.equals("<")
             || op.equals(">")
             || op.equals("<=")
             || op.equals(">=") )
      {
         result = 7;
      }
      else if ( op.equals("==")
             || op.equals("!=") )
      {
         result = 8;
      }
      else if ( op.equals("&&") )
      {
         result = 11;
      }
      else if ( op.equals("||") )
      {
         result = 12;
      }
      else if ( op.equals("if") )
      {
         result = 14;
      }
      else if ( op.equals("while") )
      {
         result = 14;
      }
      else if ( op.equals("var") )
      {
         result = 14;
      }
      else if ( op.equals("set") )
      {
         result = 14;
      }
      else if ( op.equals("print") )
      {
         result = 14;
      }

      return result;
   }

}//AST2infix
