/**
   This program parses a string from the following grammar
   into an Abstract Syntax Tree (AST).

       Prog ::= Exp
              | '(' 'prog' Exp+ ')'

        Exp ::= Lambda
              | Apply
              | If
              | While
              | Var
              | Set
              | Print
              | AExp
              | BExp
              | INTEGER
              | BOOLEAN
              | VARIABLE

      Lambda ::= '(' 'lambda' '(' VARIABLE* ')' Exp+ ')'   // a function literal

      Apply ::= '(' Exp Exp* ')'   // function value followed by actual parameters

         If ::= '(' 'if' Exp Exp Exp ')'

      While ::= '(' 'while' Exp Exp+ ')'

        Var ::= '(' 'var' VARIABLE Exp ')'

        Set ::= '(' 'set' VARIABLE Exp ')'

      Print ::= '(' 'print' Exp ')'

       BExp ::= '(' '||'  Exp Exp+ ')'
              | '(' '&&'  Exp Exp+ ')'
              | '(' '!'   Exp ')'
              | '(' RelOp Exp Exp ')'
              | '('  EqOp Exp Exp ')'

      RelOp ::= '<' | '>' | '<=' | '>='
       EqOp ::= '==' | '!='

       AExp ::= '(' '+' Exp Exp* ')'
              | '(' '-' Exp Exp? ')'
              | '(' '*' Exp Exp+ ')'
              | '(' '/' Exp Exp  ')'
              | '(' '%' Exp Exp  ')'
              | '(' '^' Exp Exp  ')'

    INTEGER ::= [0-9]+
    BOOLEAN ::= 'true' | 'false'
   VARIABLE ::= [a-zA-Z][a-zA-Z0-9]*
*/

public class Parse
{
   /**
      The methods parseProg(), parseExp(), parseLambda(),
      parseApply(), parseIf(), parseWhile(), parseVar(),
      parseSet(), parsePrint(), parseBExp(), and
      and parseAExp() are a recursive descent parser
      for the grammar of Language_3.
   */
   public static Tree parse(String program) throws ParseException
   {
      Tokenizer tokens = new Tokenizer(program);  // Tokenizer is defined
                                                  // in Tokenizer.java.
      Tree result = parseProg(tokens);  // Parse the tokens into an AST.

      if ( tokens.hasToken() )  // There shouldn't be any more tokens.
         throw new ParseException("unexpected input: "
                                 + tokens.peekToken()
                                 + "\n"+tokens + "\n");
      return result;
   }//parse()


   // Parse a Prog expression
   private static Tree parseProg(Tokenizer tokens) throws ParseException
   {
      Tree result = null;
      // Check which kind of Prog we have.
      if ( tokens.peekToken().equals("(")
        && tokens.remainingTokens() >= 2
        && tokens.peek2Token().equals("prog") )
      {
         tokens.match("(");
         result = new Tree( tokens.nextToken() );

         // Parse each Exp in the Prog.
         while (! tokens.peekToken().equals(")"))
         {
            result.addSubTree( parseExp(tokens) );
         }
         if (! tokens.hasToken())
            throw new ParseException("unexpected end of program"
                                    + "\n"+tokens + "\n");
         if (! tokens.match(")"))
            throw new ParseException("expected ')' found "
                                    + tokens.peekToken()
                                    + "\n"+tokens + "\n");
      }
      else
      {
         // Parse the single expression.
         result = parseExp(tokens);
      }

      return result;
   }//parseProg()


   // Parse an Exp expression
   private static Tree parseExp(Tokenizer tokens) throws ParseException
   {
      Tree result = null;

      if ( tokens.peekToken().equals("(") )
      {
         if (tokens.remainingTokens() < 4)
            throw new ParseException("invalid expression:"
                                    + "\n"+tokens + "\n");

         String token2 = tokens.peek2Token();

         if ( token2.equals("lambda") )
            result = parseLambda( tokens );
         else if ( token2.equals("if") )
            result = parseIf( tokens );
         else if ( token2.equals("while") )
            result = parseWhile( tokens );
         else if ( token2.equals("var") || token2.equals("fun") )
            result = parseVar( tokens );
         else if ( token2.equals("set") )
            result = parseSet( tokens );
         else if ( token2.equals("print") )
            result = parsePrint( tokens );
         else if ( token2.equals("&&") || token2.equals("||")
                || token2.equals("!") )
            result = parseBExp( tokens );  // boolean expression
         else if ( token2.equals("<")  || token2.equals(">")
                || token2.equals("<=") || token2.equals(">=")
                || token2.equals("==") || token2.equals("!=") )
            result = parseRExp( tokens );  // relational operator
         else if ( token2.equals("+") || token2.equals("-")
                || token2.equals("*") || token2.equals("/")
                || token2.equals("%") || token2.equals("^") )
            result = parseAExp( tokens );  // arithmetic expression
         else // must be a function application
            result = parseApply( tokens );
      }
      else if ( tokens.peekToken().equals("true")
             || tokens.peekToken().equals("false") )
      {
         result = new Tree( tokens.nextToken() ); // a boolean
      }
      else if ( tokens.peekToken().matches("^[-+]?[0-9]+$") )
      {
         result = new Tree( tokens.nextToken() ); // an integer
      }
      else if ( tokens.peekToken().matches("^[a-zA-Z][_a-zA-Z0-9]*$") )
      {
         result = new Tree( tokens.nextToken() ); // a variable
      }
      else
         throw new ParseException("invalid expression:"
                                 + "\n"+tokens + "\n");
      return result;
   }//parseExp()


   // Parse a Lambda expression (a function literal).
   private static Tree parseLambda(Tokenizer tokens) throws ParseException
   {
      if (tokens.remainingTokens() < 8)
         throw new ParseException("invalid function definition:"
                                 + "\n"+tokens + "\n");

      tokens.match("(");
      Tree result = new Tree( tokens.nextToken() );

      if (! tokens.match("("))
         throw new ParseException("invalid function definition:"
                                 + "\n"+tokens + "\n");
      // create a sub tree to hold the formal parameters
      Tree function = new Tree( "params" );
      // get the formal parameter names
      while ( tokens.hasToken()
           && ! tokens.peekToken().equals(")"))
      {
         // get a formal parameter name
         function.addSubTree( new Tree(tokens.nextToken()) );
      }
      tokens.match(")");
      result.addSubTree( function );

      // the function body must have at least one expression
      result.addSubTree( parseExp(tokens) );

      // get the remaining expressions in the body of the function
      while ( tokens.hasToken()
           && ! tokens.peekToken().equals(")"))
      {
         // get the next expression
         result.addSubTree( parseExp(tokens) );
      }

      if (! tokens.hasToken())
         throw new ParseException("unexpected end of program"
                                 + "\n"+tokens + "\n");
      if (! tokens.match(")"))
         throw new ParseException("expected ')' found "
                                 + tokens.peekToken()
                                 + "\n"+tokens + "\n");
      return result;
   }//parseLambda()


   // Parse an Apply expression (a function application).
   private static Tree parseApply(Tokenizer tokens) throws ParseException
   {
      if (tokens.remainingTokens() < 3)
         throw new ParseException("invalid function application:"
                                 + "\n"+tokens + "\n");

      tokens.match("(");

      // create a sub tree to hold the function application
      Tree result = new Tree( "apply" );

      // get the function-valued expression that is being applied
      result.addSubTree( parseExp(tokens) );

      // get any actual parameters
      while ( tokens.hasToken()
           && ! tokens.peekToken().equals(")"))
      {
         // get the next actual parameter
         result.addSubTree( parseExp(tokens) );
      }

      if (! tokens.hasToken())
         throw new ParseException("unexpected end of program"
                                 + "\n"+tokens + "\n");
      if (! tokens.match(")"))
         throw new ParseException("expected ')' found "
                                 + tokens.peekToken()
                                 + "\n"+tokens + "\n");
      return result;
   }//parseApply()


   // Parse an If expression
   private static Tree parseIf(Tokenizer tokens) throws ParseException
   {
      if (tokens.remainingTokens() < 6)
         throw new ParseException("invalid if-expression:"
                                 + "\n"+tokens + "\n");

      tokens.match("(");
      Tree result = new Tree( tokens.nextToken() );

      // get the boolean valued expression
      result.addSubTree( parseExp(tokens) );

      // get the then-expression
      result.addSubTree( parseExp(tokens) );

      // get the else-expression
      result.addSubTree( parseExp(tokens) );

      if (! tokens.hasToken())
         throw new ParseException("unexpected end of program"
                                 + "\n"+tokens + "\n");
      if (! tokens.match(")"))
         throw new ParseException("expected ')' found "
                                 + tokens.peekToken()
                                 + "\n"+tokens + "\n");
      return result;
   }//parseIf()


   // Parse a While expression
   private static Tree parseWhile(Tokenizer tokens) throws ParseException
   {
      if (tokens.remainingTokens() < 5)
         throw new ParseException("invalid while-expression:"
                                 + "\n"+tokens + "\n");

      tokens.match("(");
      Tree result = new Tree( tokens.nextToken() );

      // get the boolean valued expression
      result.addSubTree( parseExp(tokens) );

      // the while-loop body must have at least one expression
      result.addSubTree( parseExp(tokens) );

      // get the remaining expressions in the body of the while-loop
      while ( tokens.hasToken()
           && ! tokens.peekToken().equals(")"))
      {
         // get the next expression
         result.addSubTree( parseExp(tokens) );
      }

      if (! tokens.hasToken())
         throw new ParseException("unexpected end of program"
                                 + "\n"+tokens + "\n");
      if (! tokens.match(")"))
         throw new ParseException("expected ')' found "
                                 + tokens.peekToken()
                                 + "\n"+tokens + "\n");
      return result;
   }//parseWhile()


   // Parse a Var expression
   private static Tree parseVar(Tokenizer tokens) throws ParseException
   {
      if (tokens.remainingTokens() < 5)
         throw new ParseException("invalid var-expression:"
                                 + "\n"+tokens + "\n");

      tokens.match("(");
      Tree result = new Tree( tokens.nextToken() );

      // get the variable
      result.addSubTree( parseExp(tokens) );

      // get the expression
      result.addSubTree( parseExp(tokens) );

      if (! tokens.hasToken())
         throw new ParseException("unexpected end of program"
                                 + "\n"+tokens + "\n");
      if (! tokens.match(")"))
         throw new ParseException("expected ')' found "
                                 + tokens.peekToken()
                                 + "\n"+tokens + "\n");
      return result;
   }//parseVar()


   // Parse a Set expression
   private static Tree parseSet(Tokenizer tokens) throws ParseException
   {
      if (tokens.remainingTokens() < 5)
         throw new ParseException("invalid set-expression:"
                                 + "\n"+tokens + "\n");

      tokens.match("(");
      Tree result = new Tree( tokens.nextToken() );

      // get the variable
      result.addSubTree( parseExp(tokens) );

      // get the expression
      result.addSubTree( parseExp(tokens) );

      if (! tokens.hasToken())
         throw new ParseException("unexpected end of program"
                                 + "\n"+tokens + "\n");
      if (! tokens.match(")"))
         throw new ParseException("expected ')' found "
                                 + tokens.peekToken()
                                 + "\n"+tokens + "\n");
      return result;
   }//parseSet()


   // Parse a Print expression
   private static Tree parsePrint(Tokenizer tokens) throws ParseException
   {
      if (tokens.remainingTokens() < 4)
         throw new ParseException("invalid print-expression:"
                                 + "\n"+tokens + "\n");

      tokens.match("(");
      Tree result = new Tree( tokens.nextToken() );

      // get the expression
      result.addSubTree( parseExp(tokens) );

      if (! tokens.hasToken())
         throw new ParseException("unexpected end of program"
                                 + "\n"+tokens + "\n");
      if (! tokens.match(")"))
         throw new ParseException("expected ')' found "
                                 + tokens.peekToken()
                                 + "\n"+tokens + "\n");
      return result;
   }//parsePrint()


   // Parse a BExp (boolean) expression
   private static Tree parseBExp(Tokenizer tokens) throws ParseException
   {
      if (tokens.remainingTokens() < 4)
         throw new ParseException("invalid boolean-expression:"
                                 + "\n"+tokens + "\n");

      tokens.match("(");
      String operator = tokens.nextToken();
      Tree result = new Tree( operator );

      if (operator.equals("&&") || operator.equals("||"))
      {
         // get the first operand
         result.addSubTree( parseExp(tokens) );

         // get the second operand
         result.addSubTree( parseExp(tokens) );

         // get any remaining operands
         while ( tokens.hasToken()
              && ! tokens.peekToken().equals(")"))
         {
            // get the next operand
            result.addSubTree( parseExp(tokens) );
         }
      }
      else if (operator.equals("!"))
      {
         // get the operand
         result.addSubTree( parseExp(tokens) );
      }

      if (! tokens.hasToken())
         throw new ParseException("unexpected end of program"
                                 + "\n"+tokens + "\n");
      if (! tokens.match(")"))
         throw new ParseException("expected ')' found "
                                 + tokens.peekToken()
                                 + "\n"+tokens + "\n");
      return result;
   }//parseBExp()


   // Parse a RExp (relational) expression (which is a kind of boolean expression)
   private static Tree parseRExp(Tokenizer tokens) throws ParseException
   {
      if (tokens.remainingTokens() < 5)
         throw new ParseException("invalid relational-expression:"
                                 + "\n"+tokens + "\n");

      tokens.match("(");

      // get the operator
      Tree result = new Tree( tokens.nextToken() );
      // get the first operand
      result.addSubTree( parseExp(tokens) );
      // get the second operand
      result.addSubTree( parseExp(tokens) );

      if (! tokens.hasToken())
         throw new ParseException("unexpected end of program"
                                 + "\n"+tokens + "\n");
      if (! tokens.match(")"))
         throw new ParseException("expected ')' found "
                                 + tokens.peekToken()
                                 + "\n"+tokens + "\n");
      return result;
   }//parseRExp()


   // Parse an AExp (arithmetic) expression
   private static Tree parseAExp(Tokenizer tokens) throws ParseException
   {
      if (tokens.remainingTokens() < 4)
         throw new ParseException("invalid arithmetic-expression:"
                                 + "\n"+tokens + "\n");

      tokens.match("(");
      String operator = tokens.nextToken();
      Tree result = new Tree( operator );

      if (operator.equals("+"))
      {
         // get the first operand
         result.addSubTree( parseExp(tokens) );

         // get any remaining operands
         while ( tokens.hasToken()
              && ! tokens.peekToken().equals(")"))
         {
            // get the next operand
            result.addSubTree( parseExp(tokens) );
         }
      }
      else if (operator.equals("-"))
      {
         // get the first operand
         result.addSubTree( parseExp(tokens) );

         // get an optional second operand
         if ( tokens.hasToken()
              && ! tokens.peekToken().equals(")"))
         {
            // get the second operand
            result.addSubTree( parseExp(tokens) );
         }
      }
      else if (operator.equals("*"))
      {
         // get the first operand
         result.addSubTree( parseExp(tokens) );

         // get the second operand
         result.addSubTree( parseExp(tokens) );

         // get any remaining operands
         while ( tokens.hasToken()
              && ! tokens.peekToken().equals(")"))
         {
            // get the next operand
            result.addSubTree( parseExp(tokens) );
         }
      }
      else if (operator.equals("/")
            || operator.equals("%")
            || operator.equals("^"))
      {
         // get the first operand
         result.addSubTree( parseExp(tokens) );

         // get the second operand
         result.addSubTree( parseExp(tokens) );
      }

      if (! tokens.hasToken())
         throw new ParseException("unexpected end of program"
                                 + "\n"+tokens + "\n");
      if (! tokens.match(")"))
         throw new ParseException("expected ')' found "
                                 + tokens.peekToken()
                                 + "\n"+tokens + "\n");
      return result;
   }//parseAExp()

}//Parse
